from __future__ import unicode_literals

import logging
import pkg_resources

from .recipe import Recipe


ENTRY_POINT = "zoonado.recipes"

log = logging.getLogger(__name__)


class RecipeClassProxy(object):

    def __init__(self, client, recipe_class):
        self.client = client
        self.recipe_class = recipe_class

    def __call__(self, *args, **kwargs):
        recipe = self.recipe_class(*args, **kwargs)
        recipe.set_client(self.client)
        return recipe


class RecipeProxy(object):

    def __init__(self, client):
        self.client = client

        self.installed_classes = {}
        self.gather_installed_classes()

    def __getattr__(self, name):
        if name not in self.installed_classes:
            raise AttributeError("No such recipe: %s" % name)

        return RecipeClassProxy(self.client, self.installed_classes[name])

    def gather_installed_classes(self):
        for entry_point in pkg_resources.iter_entry_points(ENTRY_POINT):
            try:
                recipe_class = entry_point.load()
            except ImportError as e:
                log.error(
                    "Could not load recipe %s: %s", entry_point.name, str(e)
                )
                continue

            if not issubclass(recipe_class, Recipe):
                log.error(
                    "Could not load recipe %s: not a Recipe subclass",
                    entry_point.name
                )
                continue

            if not recipe_class.validate_dependencies():
                log.error(
                    "Could not load recipe %s: %s has unmet dependencies",
                    entry_point.name, recipe_class.__name__
                )
                continue

            log.debug("Loaded recipe %s", recipe_class.__name__)
            self.installed_classes[recipe_class.__name__] = recipe_class



from __future__ import unicode_literals

from tornado import gen

from zoonado import exc

from .base_lock import BaseLock


class Lock(BaseLock):

    @gen.coroutine
    def acquire(self, timeout=None):
        result = None
        while not result:
            try:
                result = yield self.wait_in_line("lock", timeout)
            except exc.SessionLost:
                continue

        raise gen.Return(result)



import logging
import random

from tornado import gen


log = logging.getLogger()


def arguments(parser):
    parser.add_argument(
        "--workers", "-w", type=int, default=5,
        help="Number of workers to launch."
    )
    parser.add_argument(
        "--znode-path", "-p", type=str, default="examplelock",
        help="ZNode path to use for the election."
    )


@gen.coroutine
def run(client, args):
    log.info("Launching %d workers.", args.workers)
    yield client.start()

    order = list(range(args.workers))
    random.shuffle(order)

    yield [worker(i, client, args) for i in order]

    yield client.close()


@gen.coroutine
def worker(number, client, args):
    election = client.recipes.LeaderElection(args.znode_path)

    yield election.join()

    if election.has_leadership:
        log.info("[WORKER #%d] I am the leader!", number)
    else:
        log.info("[WORKER #%d] not the leader.", number)

    yield election.resign()



from __future__ import unicode_literals

from .request import Request
from .response import Response
from .stat import Stat
from .acl import ACL
from .primitives import UString, Int, Buffer, Vector


class CreateRequest(Request):
    """
    """
    opcode = 1

    writes_data = True

    EPHEMERAL_FLAG = 1 << 0
    SEQUENTIAL_FLAG = 1 << 1
    CONTAINER_FLAG = 1 << 2

    parts = (
        ("path", UString),
        ("data", Buffer),
        ("acl", Vector.of(ACL)),
        ("flags", Int),
    )

    def set_flags(self, ephemeral=False, sequential=False, container=False):
        flags = 0
        if ephemeral:
            flags |= self.EPHEMERAL_FLAG
        if sequential:
            flags |= self.SEQUENTIAL_FLAG
        if container:
            flags |= self.CONTAINER_FLAG

        self.flags = flags


class CreateResponse(Response):
    """
    """
    opcode = 1

    parts = (
        ("path", UString),
    )


class Create2Request(CreateRequest):
    """
    """
    opcode = 15


class Create2Response(Response):
    """
    """
    opcode = 15

    parts = (
        ("path", UString),
        ("stat", Stat),
    )



import logging
import random
from tornado import gen
from zoonado import exc

log = logging.getLogger()


def arguments(parser):
    parser.add_argument(
        "--path", "-p", type=str, default="/examplewatcher",
        help="ZNode path to use for the example."
    )


def watcher_callback(new_data):
    left, right = new_data.split(":")
    log.info("Left: %s, Right: %s", left, right)


@gen.coroutine
def run(client, args):
    yield client.start()

    try:
        yield client.create(args.path)
    except exc.NodeExists:
        pass

    watcher = client.recipes.DataWatcher()

    watcher.add_callback(args.path, watcher_callback)

    choices = ["foo:bar", "bwee:bwoo", "derp:hork"]

    for _ in range(5):
        yield client.set_data(args.path, data=random.choice(choices))
        yield gen.sleep(1)



import logging
import random

from tornado import gen


log = logging.getLogger()


def arguments(parser):
    parser.add_argument(
        "--workers", "-w", type=int, default=5,
        help="Number of workers to launch."
    )
    parser.add_argument(
        "--znode-path", "-p", type=str, default="examplelock",
        help="ZNode path to use for the election."
    )


@gen.coroutine
def run(client, args):
    log.info("Launching %d workers.", args.workers)
    yield client.start()

    yield [
        worker(i, client, args)
        for i in range(args.workers)
    ]

    yield client.close()


@gen.coroutine
def worker(number, client, args):
    party = client.recipes.Party(args.znode_path, "worker_%d" % number)

    log.info("[WORKER #%d] Joining the party", number)

    yield party.join()

    for _ in range(10):
        log.info("[WORKER #%d] Members I see: %s", number, party.members)
        yield gen.sleep(.5)
        should_leave = random.choice([False, False, True])
        if should_leave:
            log.info("[WORKER #%d] Leaving the party temporarily", number)
            yield party.leave()
            yield gen.sleep(1)
            log.info("[WORKER #%d] Rejoining the party", number)
            yield party.join()

    yield party.leave()



from __future__ import unicode_literals

from tornado import gen
from zoonado import WatchEvent

from .base_watcher import BaseWatcher


class ChildrenWatcher(BaseWatcher):

    watched_event = WatchEvent.CHILDREN_CHANGED

    @gen.coroutine
    def fetch(self, path):
        children = yield self.client.get_children(path=path, watch=True)
        raise gen.Return(children)



import collections
import itertools
import logging
import random

from tornado import gen, ioloop
from zoonado.recipes.allocator import round_robin


log = logging.getLogger(__name__)


ANIMALS = ["cat", "dog", "mouse", "human"]


def arguments(parser):
    parser.add_argument(
        "znode", type=str,
        help="Path of the base znode to use."
    )
    parser.add_argument(
        "--workers", "-w", type=int, default=5,
        help="Number of worker coroutines to launch."
    )
    parser.add_argument(
        "--items", "-n", type=int, default=17,
        help="Number of items to allocate amongst the workers."
    )
    parser.add_argument(
        "--alloc-func", "-a", default="round_robin",
        choices=allocation_functions.keys(),
        help="Which allocation function to use."
    )


@gen.coroutine
def run(client, args):
    items = set([
        "%s::%s" % (i, random.choice(ANIMALS))
        for i in range(args.items)
    ])
    allocation_function = allocation_functions[args.alloc_func]

    yield client.start()

    for i in range(args.workers):
        ioloop.IOLoop.current().add_callback(
            worker, i, client, args.znode, allocation_function, items
        )

    yield gen.sleep(10)


@gen.coroutine
def worker(number, client, znode_path, allocation_fn, items):
    name = "worker_%s" % number

    allocator = client.recipes.Allocator(znode_path, name, allocation_fn)

    yield allocator.start()

    yield allocator.update(items)

    while True:
        log.info("[WORKER %d] My set: %s", number, allocator.allocation)
        yield gen.sleep(2)


def animal_buckets(members, items):
    animal_assignment = {}
    for member, animal in zip(itertools.cycle(members), ANIMALS):
        animal_assignment[animal] = member

    allocation = collections.defaultdict(set)
    for member, item in zip(itertools.cycle(members), items):
        animal = item.split("::")[1]
        allocation[animal_assignment[animal]].add(item)

    return allocation


allocation_functions = {
    "round_robin": round_robin,
    "buckets": animal_buckets,
}



from __future__ import unicode_literals

import operator

from .primitives import Primitive


class Part(object):
    """
    Composable building block used to define Zookeeper protocol parts.

    Behaves much like the `Primitive` class but has named "sub parts"
    stored in a ``parts`` class attribute, that can hold any `Part` or
    `Primitive` subclass.
    """
    parts = ()

    def __init__(self, **kwargs):
        part_names = set([item[0] for item in self.parts])

        for name, value in kwargs.items():
            if name not in part_names:
                raise ValueError("Unknown part name: '%s'" % name)
            part_names.discard(name)

            setattr(self, name, value)

        for name in part_names:
            setattr(self, name, None)

    def render(self, parts=None):
        """
        Returns a two-element tuple with the ``struct`` format and values.

        Iterates over the applicable sub-parts and calls `render()` on them,
        accumulating the format string and values.

        Optionally takes a subset of parts to render, default behavior is to
        render all sub-parts belonging to the class.
        """
        if not parts:
            parts = self.parts

        fmt = []
        data = []

        for name, part_class in parts:
            if issubclass(part_class, Primitive):
                part = part_class(getattr(self, name, None))
            else:
                part = getattr(self, name, None)

            part_format, part_data = part.render()

            fmt.extend(part_format)
            data.extend(part_data)

        return "".join(fmt), data

    @classmethod
    def parse(cls, buff, offset):
        """
        Given a buffer and offset, returns the parsed value and new offset.

        Calls `parse()` on the given buffer for each sub-part in order and
        creates a new instance with the results.
        """
        values = {}

        for name, part in cls.parts:
            value, new_offset = part.parse(buff, offset)

            values[name] = value
            offset = new_offset

        return cls(**values), offset

    def __eq__(self, other):
        """
        `Part` instances are equal if all of their sub-parts are also equal.
        """
        return all([
            getattr(self, part_name) == getattr(other, part_name)
            for part_name, part_class in self.parts
        ])

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):

        def subpart_string(part_info):
            part_name, part_class = part_info

            if not part_class.__name__.startswith("VectorOf"):
                return "%s=%s" % (part_name, getattr(self, part_name, None))

            return "%s=[%s]" % (
                part_name,
                ", ".join([
                    str(item) for item in getattr(self, part_name, [])
                ])
            )

        return "%s(%s)" % (
            self.__class__.__name__,
            ", ".join([
                subpart_string(part)
                for part in sorted(self.parts, key=operator.itemgetter(0))
            ])
        )



from __future__ import unicode_literals

import collections
import logging

from tornado import concurrent

from .iterables import drain


log = logging.getLogger(__name__)


class States(object):

    CONNECTED = "connected"
    SUSPENDED = "suspended"
    READ_ONLY = "read_only"
    LOST = "lost"


class SessionStateMachine(object):

    valid_transitions = set([
        (States.LOST, States.CONNECTED),
        (States.LOST, States.READ_ONLY),
        (States.CONNECTED, States.SUSPENDED),
        (States.CONNECTED, States.LOST),
        (States.READ_ONLY, States.CONNECTED),
        (States.READ_ONLY, States.SUSPENDED),
        (States.READ_ONLY, States.LOST),
        (States.SUSPENDED, States.CONNECTED),
        (States.SUSPENDED, States.READ_ONLY),
        (States.SUSPENDED, States.LOST),
    ])

    def __init__(self):
        self.current_state = States.LOST
        self.futures = collections.defaultdict(set)

    def transition_to(self, state):
        if (self.current_state, state) not in self.valid_transitions:
            raise RuntimeError(
                "Invalid session state transition: %s -> %s" % (
                    self.current_state, state
                )
            )

        log.debug("Session transition: %s -> %s", self.current_state, state)

        self.current_state = state

        for future in drain(self.futures[state]):
            if not future.done():
                future.set_result(None)

    def wait_for(self, *states):
        f = concurrent.Future()

        if self.current_state in states:
            f.set_result(None)
        else:
            for state in states:
                self.futures[state].add(f)

        return f

    def __eq__(self, state):
        return self.current_state == state

    def __ne__(self, state):
        return self.current_state != state



from __future__ import unicode_literals

from .request import Request
from .response import Response
from .primitives import Int, Long, Vector, UString


WATCH_XID = -1


class WatchEvent(Response):
    """
    """

    CREATED = 1
    DELETED = 2
    DATA_CHANGED = 3
    CHILDREN_CHANGED = 4

    DISCONNECTED = 0
    CONNECTED = 3
    AUTH_FAILED = 4
    CONNECTED_READ_ONLY = 5
    SASL_AUTHENTICATED = 6
    SESSION_EXPIRED = -112

    parts = (
        ("type", Int),
        ("state", Int),
        ("path", UString),
    )


class SetWatchesRequest(Request):
    """
    """
    opcode = 101

    parts = (
        ("relative_zxid", Long),
        ("data_watches", Vector.of(UString)),
        ("exist_watches", Vector.of(UString)),
        ("child_watches", Vector.of(UString)),
    )


class SetWatchesResponse(Response):
    """
    """
    opcode = 101

    parts = ()


class CheckWatchesRequest(Request):
    """
    """
    opcode = 17

    parts = (
        ("path", UString),
        ("type", Int),
    )


class CheckWatchesResponse(Response):
    """
    """
    opcode = 17

    parts = ()


class RemoveWatchesRequest(Request):
    """
    """
    opcode = 18

    parts = (
        ("path", UString),
        ("type", Int),
    )


class RemoveWatchesResponse(Response):
    """
    """
    opcode = 18

    parts = ()



from __future__ import unicode_literals

from .request import Request
from .response import Response
from .stat import Stat
from .primitives import Bool, UString, Vector


class GetChildrenRequest(Request):
    """
    """
    opcode = 8

    parts = (
        ("path", UString),
        ("watch", Bool),
    )


class GetChildrenResponse(Response):
    """
    """
    opcode = 8

    parts = (
        ("children", Vector.of(UString)),
    )


class GetChildren2Request(Request):
    """
    """
    opcode = 12

    parts = (
        ("path", UString),
        ("watch", Bool),
    )


class GetChildren2Response(Response):
    """
    """
    opcode = 12

    parts = (
        ("children", Vector.of(UString)),
        ("stat", Stat),
    )



from __future__ import unicode_literals


def add_metaclass(metaclass):
    """
    Class decorator for creating a class with a metaclass.

    Adapted from the six project:

    https://pythonhosted.org/six/
    """
    vars_to_skip = ('__dict__', '__weakref__')

    def wrapper(cls):
        copied_dict = {
            key: value
            for key, value in cls.__dict__.items()
            if key not in vars_to_skip
        }
        return metaclass(cls.__name__, cls.__bases__, copied_dict)
    return wrapper



from __future__ import unicode_literals

import collections
import logging
import re
import struct
import sys

from tornado import ioloop, iostream, gen, concurrent, tcpclient

from zoonado import protocol, iterables, exc


version_regex = re.compile(r'Zookeeper version: (\d)\.(\d)\.(\d)-.*')

# all requests and responses are prefixed with a 32-bit int denoting size
size_struct = struct.Struct("!i")
# replies are prefixed with an xid, zxid and error code
reply_header_struct = struct.Struct("!iqi")

log = logging.getLogger(__name__)
payload_log = logging.getLogger(__name__ + ".payload")
if payload_log.level == logging.NOTSET:
    payload_log.setLevel(logging.INFO)


class Connection(object):

    def __init__(self, host, port, watch_handler):
        self.host = host
        self.port = int(port)

        self.stream = None
        self.closing = False

        self.version_info = None
        self.start_read_only = None

        self.watch_handler = watch_handler

        self.opcode_xref = {}

        self.pending = {}
        self.pending_specials = collections.defaultdict(list)

        self.watches = collections.defaultdict(list)

    @gen.coroutine
    def connect(self):
        client = tcpclient.TCPClient()

        log.debug("Initial connection to server %s:%d", self.host, self.port)
        stream = yield client.connect(self.host, self.port)

        log.debug("Sending 'srvr' command to %s:%d", self.host, self.port)
        yield stream.write(b"srvr")
        answer = yield stream.read_until_close()

        answer = answer.decode("utf8")

        version_line = answer.split("\n")[0]
        self.version_info = tuple(
            map(int, version_regex.match(version_line).groups())
        )
        self.start_read_only = bool("READ_ONLY" in answer)

        log.debug("Version info: %s", self.version_info)
        log.debug("Read-only mode: %s", self.start_read_only)

        log.debug("Actual connection to server %s:%d", self.host, self.port)
        self.stream = yield client.connect(self.host, self.port)

    @gen.coroutine
    def send_connect(self, request):
        # meant to be used before the read_loop starts
        payload_log.debug("[SEND] (initial) %s", request)

        payload = request.serialize()
        payload = size_struct.pack(len(payload)) + payload

        yield self.stream.write(payload)

        try:
            _, zxid, response = yield self.read_response(initial_connect=True)
        except Exception:
            log.exception("Error reading connect response.")
            return

        payload_log.debug("[RECV] (initial) %s", response)

        raise gen.Return((zxid, response))

    def start_read_loop(self):
        ioloop.IOLoop.current().add_callback(self.read_loop)

    def send(self, request, xid=None):
        f = concurrent.Future()

        if self.closing:
            f.set_exception(exc.ConnectError(self.host, self.port))
            return f

        if request.special_xid:
            xid = request.special_xid

        payload_log.debug("[SEND] (xid: %s) %s", xid, request)

        payload = request.serialize(xid)
        payload = size_struct.pack(len(payload)) + payload

        self.opcode_xref[xid] = request.opcode

        if xid in protocol.SPECIAL_XIDS:
            self.pending_specials[xid].append(f)
        else:
            self.pending[xid] = f

        def handle_write(write_future):
            try:
                write_future.result()
            except Exception:
                self.abort()

        try:
            self.stream.write(payload).add_done_callback(handle_write)
        except Exception:
            self.abort()

        return f

    @gen.coroutine
    def read_loop(self):
        """
        Infinite loop that reads messages off of the socket while not closed.

        When a message is received its corresponding pending Future is set
        to have the message as its result.

        This is never used directly and is fired as a separate callback on the
        I/O loop via the `connect()` method.
        """
        while not self.closing:
            try:
                xid, zxid, response = yield self.read_response()
            except iostream.StreamClosedError:
                return
            except Exception:
                log.exception("Error reading response.")
                self.abort()
                return

            payload_log.debug("[RECV] (xid: %s) %s", xid, response)

            if xid == protocol.WATCH_XID:
                self.watch_handler(response)
                continue
            elif xid in protocol.SPECIAL_XIDS:
                f = self.pending_specials[xid].pop()
            else:
                f = self.pending.pop(xid)

            if isinstance(response, Exception):
                f.set_exception(response)
            else:
                f.set_result((zxid, response))

    @gen.coroutine
    def read_response(self, initial_connect=False):
        raw_size = yield self.stream.read_bytes(size_struct.size)
        size = size_struct.unpack(raw_size)[0]

        # connect and close op replies don't contain a reply header
        if initial_connect or self.pending_specials[protocol.CLOSE_XID]:
            raw_payload = yield self.stream.read_bytes(size)
            response = protocol.ConnectResponse.deserialize(raw_payload)
            raise gen.Return((None, None, response))

        raw_header = yield self.stream.read_bytes(reply_header_struct.size)
        xid, zxid, error_code = reply_header_struct.unpack_from(raw_header)

        if error_code:
            raise gen.Return((xid, zxid, exc.get_response_error(error_code)))

        size -= reply_header_struct.size

        raw_payload = yield self.stream.read_bytes(size)

        if xid == protocol.WATCH_XID:
            response = protocol.WatchEvent.deserialize(raw_payload)
        else:
            opcode = self.opcode_xref.pop(xid)
            response = protocol.response_xref[opcode].deserialize(raw_payload)

        raise gen.Return((xid, zxid, response))

    def abort(self, exception=exc.ConnectError):
        """
        Aborts a connection and puts all pending futures into an error state.

        If ``sys.exc_info()`` is set (i.e. this is being called in an exception
        handler) then pending futures will have that exc info set.  Otherwise
        the given ``exception`` parameter is used (defaults to
        ``ConnectError``).
        """
        log.warn("Aborting connection to %s:%s", self.host, self.port)

        def abort_pending(f):
            exc_info = sys.exc_info()
            if any(exc_info):
                f.set_exc_info(exc_info)
            else:
                f.set_exception(exception(self.host, self.port))

        for pending in self.drain_all_pending():
            abort_pending(pending)

    def drain_all_pending(self):
        for special_xid in protocol.SPECIAL_XIDS:
            for f in iterables.drain(self.pending_specials[special_xid]):
                yield f
        for _, f in iterables.drain(self.pending):
            yield f

    @gen.coroutine
    def close(self, timeout):
        if self.closing:
            return

        self.closing = True

        pending_with_timeouts = []
        for pending in self.drain_all_pending():
            pending_with_timeouts.append(gen.with_timeout(timeout, pending))

        try:
            yield list(pending_with_timeouts)
        except gen.TimeoutError:
            yield self.abort(exception=exc.TimeoutError)
        finally:
            self.stream.close()



from __future__ import unicode_literals

from .request import Request
from .response import Response
from .primitives import Int, Long, Buffer, Bool


class ConnectRequest(Request):
    """
    """
    parts = (
        ("protocol_version", Int),
        ("last_seen_zxid", Long),
        ("timeout", Int),
        ("session_id", Long),
        ("password", Buffer),
        ("read_only", Bool),
    )


class ConnectResponse(Response):
    """
    """
    parts = (
        ("protocol_version", Int),
        ("timeout", Int),
        ("session_id", Long),
        ("password", Buffer),
    )



from __future__ import unicode_literals

import logging
import time

from tornado import gen, ioloop
from zoonado import exc

from .sequential import SequentialRecipe


log = logging.getLogger(__name__)


class Lease(SequentialRecipe):

    def __init__(self, base_path, limit=1):
        super(Lease, self).__init__(base_path)
        self.limit = limit

    @gen.coroutine
    def obtain(self, duration):
        lessees = yield self.client.get_children(self.base_path)

        if len(lessees) >= self.limit:
            raise gen.Return(False)

        time_limit = time.time() + duration.total_seconds()

        try:
            yield self.create_unique_znode("lease", data=str(time_limit))
        except exc.NodeExists:
            log.warn("Lease for %s already obtained.", self.base_path)

        ioloop.IOLoop.current().call_at(time_limit, self.release)

        raise gen.Return(True)

    @gen.coroutine
    def release(self):
        yield self.delete_unique_znode("lease")



from __future__ import unicode_literals

import logging
import re
import uuid

from tornado import gen
from zoonado import exc, WatchEvent

from .recipe import Recipe


log = logging.getLogger(__name__)

sequential_re = re.compile(r'.*[0-9]{10}$')


class SequentialRecipe(Recipe):

    def __init__(self, base_path):
        super(SequentialRecipe, self).__init__(base_path)
        self.guid = uuid.uuid4().hex

        self.owned_paths = {}

    def sequence_number(self, sibling):
        return int(sibling[-10:])

    def determine_znode_label(self, sibling):
        return sibling.rsplit("-", 2)[0]

    def sibling_path(self, path):
        return "/".join([self.base_path, path])

    @gen.coroutine
    def create_unique_znode(self, znode_label, data=None):
        path = self.sibling_path(znode_label + "-" + self.guid + "-")

        try:
            created_path = yield self.client.create(
                path, data=data, ephemeral=True, sequential=True
            )
        except exc.NoNode:
            yield self.ensure_path()
            created_path = yield self.client.create(
                path, data=data, ephemeral=True, sequential=True
            )

        self.owned_paths[znode_label] = created_path

    @gen.coroutine
    def delete_unique_znode(self, znode_label):
        try:
            yield self.client.delete(self.owned_paths[znode_label])
        except exc.NoNode:
            pass

    @gen.coroutine
    def analyze_siblings(self):
        siblings = yield self.client.get_children(self.base_path)
        siblings = [name for name in siblings if sequential_re.match(name)]

        siblings.sort(key=self.sequence_number)

        owned_positions = {}

        for index, path in enumerate(siblings):
            if self.guid in path:
                owned_positions[self.determine_znode_label(path)] = index

        raise gen.Return((owned_positions, siblings))

    @gen.coroutine
    def wait_on_sibling(self, sibling, time_limit=None):
        log.debug("Waiting on sibling %s", sibling)

        path = self.sibling_path(sibling)

        unblocked = self.client.wait_for_event(WatchEvent.DELETED, path)
        if time_limit:
            unblocked = gen.with_timeout(time_limit, unblocked)

        exists = yield self.client.exists(path=path, watch=True)
        if not exists:
            unblocked.set_result(None)

        try:
            yield unblocked
        except gen.TimeoutError:
            raise exc.TimeoutError



from __future__ import unicode_literals

import collections
import logging

from tornado import ioloop, gen

from zoonado import exc

from .recipe import Recipe


log = logging.getLogger(__name__)


class BaseWatcher(Recipe):

    watched_event = None

    def __init__(self, *args, **kwargs):
        super(BaseWatcher, self).__init__(*args, **kwargs)
        self.callbacks = collections.defaultdict(set)

    def add_callback(self, path, callback):
        self.callbacks[path].add(callback)

        if len(self.callbacks[path]) == 1:
            ioloop.IOLoop.current().add_callback(self.watch_loop, path)

    def remove_callback(self, path, callback):
        self.callbacks[path].discard(callback)

    @gen.coroutine
    def fetch(self, path):
        raise NotImplementedError

    @gen.coroutine
    def watch_loop(self, path):
        while self.callbacks[path]:
            wait = self.client.wait_for_event(self.watched_event, path)

            log.debug("Fetching data for %s", path)
            try:
                result = yield self.fetch(path)
            except exc.NoNode:
                return

            yield wait

            for callback in self.callbacks[path]:
                ioloop.IOLoop.current().add_callback(callback, result)



import datetime
import logging
import random

from tornado import gen, ioloop
from zoonado import exc


log = logging.getLogger()


def arguments(parser):
    parser.add_argument(
        "--path", "-p", type=str, default="/example-task-lease"
    )
    parser.add_argument(
        "--limit", "-l", type=int, default=1,
        help="Max number of simultaneous leases."
    )


@gen.coroutine
def run(client, args):
    yield client.start()

    try:
        yield client.create(args.path)
    except exc.NodeExists:
        pass

    # simulate a cron job, same task fired at an interval
    for i in range(8):
        ioloop.IOLoop.current().add_callback(limited_task, i, client, args)
        yield gen.sleep(1)


@gen.coroutine
def limited_task(number, client, args):
    lease = client.recipes.Lease(args.path, args.limit)

    seconds = random.choice([1, 2, 3])

    obtained = yield lease.obtain(duration=datetime.timedelta(seconds=seconds))
    if obtained:
        log.info("[ITERATION #%d] Got lease for %d seconds", number, seconds)
    else:
        log.info("[ITERATION #%d] No lease available, can't work", number)



import logging

from tornado import ioloop, gen
from zoonado import exc


log = logging.getLogger()


def arguments(_):
    pass


@gen.coroutine
def run(client, args):
    config_path = "/exampleconfig"
    loop = ioloop.IOLoop.current()

    yield client.start()

    config = client.recipes.TreeCache(config_path)

    yield config.start()

    try:
        yield client.create(config_path + "/running", data="yes")
    except exc.NodeExists:
        yield client.set_data(config_path + "/running", data="yes")

    for path in ["foo", "bar", "bazz", "bloo"]:
        try:
            yield client.create(config_path + "/" + path, data="1")
        except exc.NodeExists:
            yield client.set_data(config_path + "/" + path, data="1")

    loop.add_callback(foo, config)
    loop.add_callback(bar, config)
    loop.add_callback(bazz, config)
    loop.add_callback(bloo, config)

    yield gen.sleep(1)

    yield client.set_data(config_path + "/foo", "3")

    yield gen.sleep(1)

    yield client.set_data(config_path + "/bar", "2")

    yield client.set_data(config_path + "/bazz", "5")

    yield gen.sleep(6)

    yield client.set_data(config_path + "/running", data="no")

    yield gen.sleep(2)

    yield client.close()


@gen.coroutine
def foo(config):
    while config.running.value == "yes":
        log.info("[FOO] doing work for %s seconds!", config.foo.value)
        yield gen.sleep(int(config.foo.value))

    log.info("[FOO] no longer working.")


@gen.coroutine
def bar(config):
    while config.running.value == "yes":
        log.info("[BAR] doing work for %s seconds!", config.bar.value)
        yield gen.sleep(int(config.bar.value))

    log.info("[BAR] no longer working.")


@gen.coroutine
def bazz(config):
    while config.running.value == "yes":
        log.info("[BAZZ] doing work for %s seconds!", config.bazz.value)
        yield gen.sleep(int(config.bazz.value))

    log.info("[BAZZ] no longer working.")


@gen.coroutine
def bloo(config):
    while config.running.value == "yes":
        log.info("[BLOO] doing work for %s seconds!", config.bloo.value)
        yield gen.sleep(int(config.bloo.value))

    log.info("[BLOO] no longer working.")



import logging
import random
from tornado import gen
from zoonado import exc

log = logging.getLogger()


def arguments(parser):
    parser.add_argument(
        "--path", "-p", type=str, default="/examplewatcher",
        help="ZNode path to use for the example."
    )


def watcher_callback(children):
    children.sort()
    log.info("There are %d items now: %s", len(children), children)


@gen.coroutine
def run(client, args):
    yield client.start()

    try:
        yield client.create(args.path)
    except exc.NodeExists:
        pass

    watcher = client.recipes.ChildrenWatcher()

    watcher.add_callback(args.path, watcher_callback)

    to_make = ["cat", "dog", "mouse", "human"]
    random.shuffle(to_make)

    for item in to_make:
        yield client.create(args.path + "/" + item, ephemeral=True)
        yield gen.sleep(1)

    for item in to_make:
        yield client.delete(args.path + "/" + item)



from __future__ import unicode_literals

import time

from tornado import gen

from zoonado import exc, WatchEvent

from .recipe import Recipe


class Barrier(Recipe):

    def __init__(self, path):
        super(Barrier, self).__init__()
        self.path = path

    @gen.coroutine
    def create(self):
        yield self.ensure_path()

    @gen.coroutine
    def lift(self):
        try:
            yield self.client.delete(self.path)
        except exc.NoNode:
            pass

    @gen.coroutine
    def wait(self, timeout=None):
        time_limit = None
        if timeout is not None:
            time_limit = time.time() + timeout

        barrier_lifted = self.client.wait_for_event(
            WatchEvent.DELETED, self.path
        )

        if time_limit:
            barrier_lifted = gen.with_timeout(time_limit, barrier_lifted)

        exists = yield self.client.exists(path=self.path, watch=True)
        if not exists:
            return

        try:
            yield barrier_lifted
        except gen.TimeoutError:
            raise exc.TimeoutError



from __future__ import unicode_literals

from .request import Request
from .response import Response


PING_XID = -2


class PingRequest(Request):
    """
    """
    opcode = 11
    special_xid = PING_XID

    parts = ()


class PingResponse(Response):
    """
    """
    opcode = 11

    parts = ()



from __future__ import unicode_literals

from .request import Request
from .response import Response
from .primitives import UString, Int


class DeleteRequest(Request):
    """
    """
    opcode = 2

    writes_data = True

    parts = (
        ("path", UString),
        ("version", Int),
    )


class DeleteResponse(Response):
    """
    """
    opcode = 2

    parts = ()



from __future__ import unicode_literals

import contextlib
import logging
import time

from tornado import gen, ioloop

from zoonado import exc, states

from .sequential import SequentialRecipe


log = logging.getLogger(__name__)


class BaseLock(SequentialRecipe):

    @gen.coroutine
    def wait_in_line(self, znode_label, timeout=None, blocked_by=None):
        time_limit = None
        if timeout is not None:
            time_limit = time.time() + timeout

        yield self.create_unique_znode(znode_label)

        while True:
            if time_limit and time.time() >= time_limit:
                raise exc.TimeoutError

            owned_positions, contenders = yield self.analyze_siblings()
            if znode_label not in owned_positions:
                raise exc.SessionLost

            blockers = contenders[:owned_positions[znode_label]]
            if blocked_by:
                blockers = [
                    contender for contender in blockers
                    if self.determine_znode_label(contender) in blocked_by
                ]

            if not blockers:
                break

            yield self.wait_on_sibling(blockers[-1], time_limit)

        raise gen.Return(self.make_contextmanager(znode_label))

    def make_contextmanager(self, znode_label):
        state = {"acquired": True}

        def still_acquired():
            return state["acquired"]

        @gen.coroutine
        def handle_session_loss():
            yield self.client.session.state.wait_for(states.States.LOST)
            if not state["acquired"]:
                return

            log.warn(
                "Session expired at some point, lock %s no longer acquired.",
                self
            )
            state["acquired"] = False

        ioloop.IOLoop.current().add_callback(handle_session_loss)

        @gen.coroutine
        def on_exit():
            state["acquired"] = False
            yield self.delete_unique_znode(znode_label)

        @contextlib.contextmanager
        def context_manager():
            try:
                yield still_acquired
            finally:
                ioloop.IOLoop.current().add_callback(on_exit)

        return context_manager()


class LockLostError(exc.ZKError):
    pass



from __future__ import unicode_literals

import time

from tornado import gen, ioloop, concurrent

from .sequential import SequentialRecipe


class LeaderElection(SequentialRecipe):

    def __init__(self, base_path):
        super(LeaderElection, self).__init__(base_path)
        self.has_leadership = False

        self.leadership_future = concurrent.Future()

    @gen.coroutine
    def join(self):
        yield self.create_unique_znode("candidate")
        yield self.check_position()

    @gen.coroutine
    def check_position(self, _=None):
        owned_positions, candidates = yield self.analyze_siblings()
        if "candidate" not in owned_positions:
            return

        position = owned_positions["candidate"]

        self.has_leadership = bool(position == 0)

        if self.has_leadership:
            self.leadership_future.set_result(None)
            return

        moved_up = self.wait_on_sibling(candidates[position - 1])

        ioloop.IOLoop.current().add_future(moved_up, self.check_position)

    @gen.coroutine
    def wait_for_leadership(self, timeout=None):
        if self.has_leadership:
            return

        time_limit = None
        if timeout is not None:
            time_limit = time.time() + timeout

        if time_limit:
            yield gen.with_timeout(time_limit, self.leadership_future)
        else:
            yield self.leadership_future

    @gen.coroutine
    def resign(self):
        yield self.delete_unique_znode("candidate")



import logging
import random

from tornado import gen


log = logging.getLogger()


def arguments(parser):
    parser.add_argument(
        "--path", "-b", type=str, default="/example-barrier",
        help="ZNode path to use for the double barrier."
    )
    parser.add_argument(
        "--workers", "-w", type=int, default=5,
        help="Number of worker coroutines."
    )
    parser.add_argument(
        "--min-workers", "-m", type=int, default=3,
        help="Minimum number of workers required to lift the barrier."
    )


@gen.coroutine
def run(client, args):
    yield client.start()

    workers = []

    for i in range(args.workers):
        workers.append(worker(i, client, args))
        yield gen.sleep(1)

    yield workers

    yield client.close()


@gen.coroutine
def worker(number, client, args):
    log.info("[WORKER #%d] Starting up", number)

    barrier = client.recipes.DoubleBarrier(args.path, args.min_workers)

    log.info("[WORKER #%d] Entering barrier...", number)
    yield barrier.enter()

    workload = random.choice([3, 4, 5])

    log.info("[WORKER #%d] My workload is %d", number, workload)

    for i in range(workload):
        log.info("[WORKER #%d] Doing task %d", number, i + 1)
        yield gen.sleep(1)

    log.info("[WORKER #%d] Workload complete, leaving barrier", number)
    yield barrier.leave()

    log.info("[WORKER #%d] All done!", number)



import logging
import random
import threading

from tornado import gen, ioloop
from zoonado import Zoonado


log = logging.getLogger()

monitor_ioloop = None


def arguments(_):
    pass


@gen.coroutine
def run(client, args):
    yield client.start()

    yield client.create("/shared-znode", ephemeral=True)

    monitor_thread = threading.Thread(target=monitor_data, args=(args,))
    monitor_thread.start()

    threads = [
        threading.Thread(name="A", target=launch_loop, args=(args,)),
        threading.Thread(name="B", target=launch_loop, args=(args,)),
        threading.Thread(name="C", target=launch_loop, args=(args,)),
    ]

    for thread in threads:
        thread.start()

    for thread in threads:
        thread.join()

    monitor_ioloop.stop()

    monitor_thread.join()

    yield client.close()


def monitor_data(args):
    global monitor_ioloop

    name = threading.current_thread().name
    log.info("Launching loop in thread %s", name)

    io_loop = ioloop.IOLoop()

    io_loop.make_current()

    monitor_ioloop = io_loop

    @gen.coroutine
    def monitor():
        client = Zoonado(args.servers, chroot=args.chroot)
        yield client.start()

        def data_callback(new_data):
            log.info("Shared data set to '%s'", new_data)

        watcher = client.recipes.DataWatcher()
        watcher.add_callback("/shared-znode", data_callback)

        yield gen.moment

    io_loop.add_callback(monitor)

    io_loop.start()


def launch_loop(args):
    name = threading.current_thread().name
    log.info("Launching loop in thread %s", name)

    io_loop = ioloop.IOLoop()

    io_loop.make_current()

    io_loop.add_callback(update_loop, name, args, io_loop)

    io_loop.start()


@gen.coroutine
def update_loop(name, args, io_loop):
    log.info("[LOOP %s] starting up!", name)

    client = Zoonado(args.servers, chroot=args.chroot)

    try:
        yield client.start()

        for i in range(5):
            yield client.exists("/shared-znode")
            expected_version = client.stat_cache["/shared-znode"].version

            yield gen.sleep(random.choice([.2, .4, .5]))

            log.info(
                "[LOOP %s] I expect the shared znode to have version %s",
                name, client.stat_cache["/shared-znode"].version
            )

            txn = client.begin_transaction()

            txn.create("/znode-" + name, ephemeral=True)
            txn.check_version("/shared-znode", expected_version)
            txn.set_data("/shared-znode", "altered by loop %s!" % name)
            txn.delete("/znode-" + name)

            log.info("[LOOP %s] committing...", name)
            result = yield txn.commit()
            if not result:
                log.info("[LOOP %s] rolled back!", name)

        yield client.close()
    finally:
        io_loop.stop()



from __future__ import unicode_literals

import functools
import logging

from tornado import gen, concurrent

from zoonado import protocol, exc

from .recipes.proxy import RecipeProxy
from .session import Session
from .transaction import Transaction
from .features import Features
from .encoding import default_encoder, default_decoder


log = logging.getLogger(__name__)


def znode_method(fn):

    @functools.wraps(fn)
    def wrapper(client, path, *args, **kwargs):
        path = client.normalize_path(path)

        return fn(client, path, *args, **kwargs)

    return wrapper


class Zoonado(object):

    def __init__(
            self,
            servers,
            chroot=None,
            data_encoder=None,
            data_decoder=None,
            session_timeout=10,
            default_acl=None,
            retry_policy=None,
            allow_read_only=False,
    ):
        self.chroot = None
        if chroot:
            self.chroot = self.normalize_path(chroot)
            log.info("Using chroot '%s'", self.chroot)

        self.data_encoder = data_encoder or default_encoder
        self.data_decoder = data_decoder or default_decoder

        self.session = Session(
            servers, session_timeout, retry_policy, allow_read_only
        )

        self.default_acl = default_acl or [protocol.UNRESTRICTED_ACCESS]

        self.stat_cache = {}

        self.recipes = RecipeProxy(self)

    def normalize_path(self, path):
        if self.chroot:
            path = "/".join([self.chroot, path])

        normalized = "/".join([
            name for name in path.split("/")
            if name
        ])

        return "/" + normalized

    def denormalize_path(self, path):
        if self.chroot and path.startswith(self.chroot):
            path = path[len(self.chroot):]

        return path

    @gen.coroutine
    def start(self):
        yield self.session.start()

        if self.chroot:
            yield self.ensure_path("/")

    @property
    def features(self):
        if self.session.conn:
            return Features(self.session.conn.version_info)
        else:
            return Features((0, 0, 0))

    @gen.coroutine
    def send(self, request):
        response = yield self.session.send(request)

        if getattr(request, "path", None) and getattr(response, "stat", None):
            self.stat_cache[
                self.denormalize_path(request.path)
            ] = response.stat

        raise gen.Return(response)

    @gen.coroutine
    def close(self):
        yield self.session.close()

    def wait_for_event(self, event_type, path):
        path = self.normalize_path(path)

        f = concurrent.Future()

        def set_future(_):
            if not f.done():
                f.set_result(None)
            self.session.remove_watch_callback(event_type, path, set_future)

        self.session.add_watch_callback(event_type, path, set_future)

        return f

    @gen.coroutine
    @znode_method
    def exists(self, path, watch=False):
        try:
            yield self.send(protocol.ExistsRequest(path=path, watch=watch))
        except exc.NoNode:
            raise gen.Return(False)

        raise gen.Return(True)

    @gen.coroutine
    @znode_method
    def create(
            self, path, data=None, acl=None,
            ephemeral=False, sequential=False, container=False,
    ):
        if container and not self.features.containers:
            raise ValueError("Cannot create container, feature unavailable.")

        acl = acl or self.default_acl

        data = self.data_encoder(data)

        if self.features.create_with_stat:
            request_class = protocol.Create2Request
        else:
            request_class = protocol.CreateRequest

        request = request_class(path=path, data=data, acl=acl)
        request.set_flags(ephemeral, sequential, container)

        response = yield self.send(request)

        raise gen.Return(self.denormalize_path(response.path))

    @gen.coroutine
    @znode_method
    def ensure_path(self, path, acl=None):
        acl = acl or self.default_acl

        paths_to_make = []
        for segment in path[1:].split("/"):
            if not paths_to_make:
                paths_to_make.append("/" + segment)
                continue

            paths_to_make.append("/".join([paths_to_make[-1], segment]))

        while paths_to_make:
            path = paths_to_make[0]

            if self.features.create_with_stat:
                request = protocol.Create2Request(path=path, acl=acl)
            else:
                request = protocol.CreateRequest(path=path, acl=acl)
            request.set_flags(
                ephemeral=False, sequential=False,
                container=self.features.containers
            )

            try:
                yield self.send(request)
            except exc.NodeExists:
                pass

            paths_to_make.pop(0)

    @gen.coroutine
    @znode_method
    def delete(self, path, force=False):
        version = self.determine_znode_version(path, force)

        yield self.send(protocol.DeleteRequest(path=path, version=version))

    @gen.coroutine
    @znode_method
    def get_data(self, path, watch=False):
        response = yield self.send(
            protocol.GetDataRequest(path=path, watch=watch)
        )
        data = self.data_decoder(response.data)

        raise gen.Return(data)

    @gen.coroutine
    @znode_method
    def set_data(self, path, data, force=False):
        data = self.data_encoder(data)
        version = self.determine_znode_version(path, force)

        yield self.send(
            protocol.SetDataRequest(path=path, data=data, version=version)
        )

    @gen.coroutine
    @znode_method
    def get_children(self, path, watch=False):
        response = yield self.send(
            protocol.GetChildren2Request(path=path, watch=watch)
        )
        raise gen.Return(response.children)

    @gen.coroutine
    @znode_method
    def get_acl(self, path):
        response = yield self.send(protocol.GetACLRequest(path=path))
        raise gen.Return(response.acl)

    @gen.coroutine
    @znode_method
    def set_acl(self, path, acl, force=False):
        version = self.determine_znode_version(path, force)

        yield self.send(
            protocol.SetACLRequest(path=path, acl=acl, version=version)
        )

    def begin_transaction(self):
        return Transaction(self)

    def determine_znode_version(self, path, force):
        if not force and path in self.stat_cache:
            return self.stat_cache[path].version
        else:
            return -1



from __future__ import unicode_literals

from .request import Request
from .response import Response
from .stat import Stat
from .primitives import Bool, UString, Buffer, Int


class GetDataRequest(Request):
    """
    """
    opcode = 4

    parts = (
        ("path", UString),
        ("watch", Bool),
    )


class GetDataResponse(Response):
    """
    """
    opcode = 4

    parts = (
        ("data", Buffer),
        ("stat", Stat)
    )


class SetDataRequest(Request):
    """
    """
    opcode = 5

    writes_data = True

    parts = (
        ("path", UString),
        ("data", Buffer),
        ("version", Int),
    )


class SetDataResponse(Response):
    """
    """
    opcode = 5

    parts = (
        ("stat", Stat),
    )



import logging

import six


log = logging.getLogger(__name__)


def default_encoder(data):
    if data is None or isinstance(data, six.binary_type):
        return data

    try:
        return data.encode("utf8")
    except Exception:
        log.exception("Error encoding data: %r", data)
        raise


def default_decoder(data):
    if data is None:
        return data

    try:
        return data.decode("utf8")
    except UnicodeDecodeError:
        return data



from __future__ import unicode_literals

import logging

from tornado import gen, concurrent

from .children_watcher import ChildrenWatcher
from .sequential import SequentialRecipe


log = logging.getLogger(__name__)


class Party(SequentialRecipe):

    sub_recipes = {
        "watcher": ChildrenWatcher,
    }

    def __init__(self, base_path, name):
        super(Party, self).__init__(base_path)

        self.name = name
        self.members = []
        self.change_future = None

    @gen.coroutine
    def join(self):
        yield self.create_unique_znode(self.name)

        _, siblings = yield self.analyze_siblings()
        self.update_members(siblings)

        self.watcher.add_callback(self.base_path, self.update_members)

    @gen.coroutine
    def wait_for_change(self):
        if not self.change_future or self.change_future.done():
            self.change_future = concurrent.Future()

        yield self.change_future

    @gen.coroutine
    def leave(self):
        self.watcher.remove_callback(self.base_path, self.update_members)
        yield self.delete_unique_znode(self.name)

    def update_members(self, raw_sibling_names):
        new_members = [
            self.determine_znode_label(sibling)
            for sibling in raw_sibling_names
        ]

        self.members = new_members
        if self.change_future and not self.change_future.done():
            self.change_future.set_result(new_members)



from __future__ import unicode_literals

from tornado import gen
from zoonado import WatchEvent

from .base_watcher import BaseWatcher


class DataWatcher(BaseWatcher):

    watched_event = WatchEvent.DATA_CHANGED

    @gen.coroutine
    def fetch(self, path):
        data = yield self.client.get_data(path=path, watch=True)
        raise gen.Return(data)




version_info = (0, 9, 2)

__version__ = ".".join((str(point) for point in version_info))

try:
    import tornado  # noqa
    from .client import Zoonado  # noqa
    from .protocol import WatchEvent  # noqa
    from .protocol.acl import ACL  # noqa
    from .retry import RetryPolicy  # noqa
except ImportError:  # pragma: no cover
    pass



from setuptools import setup, find_packages

from zoonado import __version__


setup(
    name="zoonado",
    version=__version__,
    description="Async tornado client for Zookeeper.",
    author="William Glass",
    author_email="william.glass@gmail.com",
    url="http://github.com/wglass/zoonado",
    license="Apache",
    keywords=["zookeeper", "tornado", "async", "distributed"],
    packages=find_packages(exclude=["tests", "tests.*"]),
    install_requires=[
        "tornado>=4.1",
        "six"
    ],
    entry_points={
        "zoonado.recipes": [
            "data_watcher = zoonado.recipes.data_watcher:DataWatcher",
            "children_watcher" +
            " = zoonado.recipes.children_watcher:ChildrenWatcher",
            "lock = zoonado.recipes.lock:Lock",
            "shared_lock = zoonado.recipes.shared_lock:SharedLock",
            "lease = zoonado.recipes.lease:Lease",
            "barrier = zoonado.recipes.barrier:Barrier",
            "double_barrier = zoonado.recipes.double_barrier:DoubleBarrier",
            "election = zoonado.recipes.election:LeaderElection",
            "party = zoonado.recipes.party:Party",
            "counter = zoonado.recipes.counter:Counter",
            "tree_cache = zoonado.recipes.tree_cache:TreeCache",
            "allocator = zoonado.recipes.allocator:Allocator",
        ],
    },
    tests_require=[
        "nose",
        "mock",
        "coverage",
        "flake8>=3.0.0",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: Implementation",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
    ],
)



import logging
import random

from tornado import gen

log = logging.getLogger()


def arguments(parser):
    parser.add_argument(
        "--reader-count", "-r", type=int, default=6,
        help="Number of readers to launch.",
    )
    parser.add_argument(
        "--writer-count", "-w", type=int, default=2,
        help="Number of writers to launch.",
    )


@gen.coroutine
def run(client, args):
    workers = ([reader] * args.reader_count) + ([writer] * args.writer_count)
    random.shuffle(workers)

    yield client.start()

    yield [
        worker_func(i, client)
        for i, worker_func in enumerate(workers)
    ]

    yield client.close()


@gen.coroutine
def reader(number, client):
    lock = client.recipes.SharedLock("/examplelock")

    wait = random.choice([1, 2, 3])

    log.info("[READER #%d] Acquiring lock...", number)
    with (yield lock.acquire_read()):
        log.info("[READER #%d] Got lock! Sleeping %d seconds", number, wait)
        yield gen.sleep(wait)
        log.info("[READER #%d] Done!", number)


@gen.coroutine
def writer(number, client):
    lock = client.recipes.SharedLock("/examplelock")

    wait = random.choice([2, 3])

    log.info("[WRITER #%d] Acquiring lock...", number)
    with (yield lock.acquire_write()):
        log.info("[WRITER #%d] Got lock! Sleeping %d seconds", number, wait)
        yield gen.sleep(wait)
        log.info("[WRITER #%d] Done!", number)



from __future__ import unicode_literals

import collections
import logging
import random

from tornado import gen, ioloop, iostream

from zoonado import protocol, exc
from .connection import Connection
from .states import States, SessionStateMachine
from .retry import RetryPolicy


DEFAULT_ZOOKEEPER_PORT = 2181

MAX_FIND_WAIT = 60  # in seconds

HEARTBEAT_FREQUENCY = 3  # heartbeats per timeout interval


log = logging.getLogger(__name__)


class Session(object):

    def __init__(self, servers, timeout, retry_policy, allow_read_only):
        self.hosts = []
        for server in servers.split(","):
            if ":" in server:
                host, port = server.split(":")
            else:
                host = server
                port = DEFAULT_ZOOKEEPER_PORT

            self.hosts.append((host, port))

        self.conn = None
        self.state = SessionStateMachine()

        self.retry_policy = retry_policy or RetryPolicy.forever()
        self.allow_read_only = allow_read_only

        self.xid = 0
        self.last_zxid = None

        self.session_id = None
        self.timeout = timeout
        self.password = b'\x00'

        self.heartbeat_handle = None

        self.watch_callbacks = collections.defaultdict(set)

        self.closing = False

    @gen.coroutine
    def ensure_safe_state(self, writing=False):
        safe_states = [States.CONNECTED]
        if self.allow_read_only and not writing:
            safe_states.append(States.READ_ONLY)

        if self.state in safe_states:
            return

        yield self.state.wait_for(*safe_states)

    @gen.coroutine
    def start(self):
        io_loop = ioloop.IOLoop.current()
        io_loop.add_callback(self.set_heartbeat)
        io_loop.add_callback(self.repair_loop)

        yield self.ensure_safe_state()

    @gen.coroutine
    def find_server(self, allow_read_only):
        conn = None

        retry_policy = RetryPolicy.exponential_backoff(maximum=MAX_FIND_WAIT)

        while not conn:
            yield retry_policy.enforce()

            servers = random.sample(self.hosts, len(self.hosts))
            for host, port in servers:
                log.info("Connecting to %s:%s", host, port)
                conn = yield self.make_connection(host, port)
                if not conn or (conn.start_read_only and not allow_read_only):
                    continue

            if not conn:
                log.warn("No servers available, will keep trying.")

        old_conn = self.conn
        self.conn = conn

        io_loop = ioloop.IOLoop.current()

        if old_conn:
            io_loop.add_callback(old_conn.close, self.timeout)

        if conn.start_read_only:
            io_loop.add_callback(self.find_server, allow_read_only=False)

    @gen.coroutine
    def make_connection(self, host, port):
        conn = Connection(host, port, watch_handler=self.event_dispatch)
        try:
            yield conn.connect()
        except Exception:
            log.exception("Couldn't connect to %s:%s", host, port)
            return

        raise gen.Return(conn)

    @gen.coroutine
    def establish_session(self):
        log.info("Establising session.")
        zxid, response = yield self.conn.send_connect(
            protocol.ConnectRequest(
                protocol_version=0,
                last_seen_zxid=self.last_zxid or 0,
                timeout=int((self.timeout or 0) * 1000),
                session_id=self.session_id or 0,
                password=self.password,
                read_only=self.allow_read_only,
            )
        )
        self.last_zxid = zxid

        if response.session_id == 0:  # invalid session, probably expired
            self.state.transition_to(States.LOST)
            raise exc.SessionLost()

        log.info("Got session id %s", hex(response.session_id))
        log.info("Negotiated timeout: %s seconds", response.timeout / 1000)

        self.session_id = response.session_id
        self.password = response.password
        self.timeout = response.timeout / 1000

        self.last_zxid = zxid

    @gen.coroutine
    def repair_loop(self):
        while not self.closing:
            yield self.state.wait_for(States.SUSPENDED, States.LOST)
            if self.closing:
                break

            yield self.find_server(allow_read_only=self.allow_read_only)

            session_was_lost = self.state == States.LOST

            try:
                yield self.establish_session()
            except exc.SessionLost:
                self.conn.abort(exc.SessionLost)
                yield self.conn.close(self.timeout)
                self.session_id = None
                self.password = b'\x00'
                continue

            if self.conn.start_read_only:
                self.state.transition_to(States.READ_ONLY)
            else:
                self.state.transition_to(States.CONNECTED)

            self.conn.start_read_loop()

            if session_was_lost:
                yield self.set_existing_watches()

    @gen.coroutine
    def send(self, request):
        response = None
        while not response:
            yield self.retry_policy.enforce(request)
            yield self.ensure_safe_state(writing=request.writes_data)

            try:
                self.xid += 1
                zxid, response = yield self.conn.send(request, xid=self.xid)
                self.last_zxid = zxid
                self.set_heartbeat()
                self.retry_policy.clear(request)
            except exc.ConnectError:
                self.state.transition_to(States.SUSPENDED)

        raise gen.Return(response)

    def set_heartbeat(self):
        timeout = self.timeout / HEARTBEAT_FREQUENCY

        io_loop = ioloop.IOLoop.current()

        if self.heartbeat_handle:
            io_loop.remove_timeout(self.heartbeat_handle)

        self.heartbeat_handle = io_loop.call_later(timeout, self.heartbeat)

    @gen.coroutine
    def heartbeat(self):
        if self.closing:
            return
        yield self.ensure_safe_state()

        try:
            zxid, _ = yield self.conn.send(protocol.PingRequest())
            self.last_zxid = zxid
        except (exc.ConnectError, iostream.StreamClosedError):
            self.state.transition_to(States.SUSPENDED)
        finally:
            self.set_heartbeat()

    def add_watch_callback(self, event_type, path, callback):
        self.watch_callbacks[(event_type, path)].add(callback)

    def remove_watch_callback(self, event_type, path, callback):
        self.watch_callbacks[(event_type, path)].discard(callback)

    def event_dispatch(self, event):
        log.debug("Got watch event: %s", event)

        if event.type:
            key = (event.type, event.path)
            for callback in self.watch_callbacks[key]:
                ioloop.IOLoop.current().add_callback(callback, event.path)
            return

        if event.state == protocol.WatchEvent.DISCONNECTED:
            log.error("Got 'disconnected' watch event.")
            self.state.transition_to(States.LOST)
        elif event.state == protocol.WatchEvent.SESSION_EXPIRED:
            log.error("Got 'session expired' watch event.")
            self.state.transition_to(States.LOST)
        elif event.state == protocol.WatchEvent.AUTH_FAILED:
            log.error("Got 'auth failed' watch event.")
            self.state.transition_to(States.LOST)
        elif event.state == protocol.WatchEvent.CONNECTED_READ_ONLY:
            log.warn("Got 'connected read only' watch event.")
            self.state.transition_to(States.READ_ONLY)
        elif event.state == protocol.WatchEvent.SASL_AUTHENTICATED:
            log.info("Authentication successful.")
        elif event.state == protocol.WatchEvent.CONNECTED:
            log.info("Got 'connected' watch event.")
            self.state.transition_to(States.CONNECTED)

    @gen.coroutine
    def set_existing_watches(self):
        if not self.watch_callbacks:
            return

        request = protocol.SetWatchesRequest(
            relative_zxid=self.last_zxid or 0,
            data_watches=[],
            exist_watches=[],
            child_watches=[],
        )

        for event_type, path in self.watch_callbacks.keys():
            if event_type == protocol.WatchEvent.CREATED:
                request.exist_watches.append(path)
            if event_type == protocol.WatchEvent.DATA_CHANGED:
                request.data_watches.append(path)
            elif event_type == protocol.WatchEvent.CHILDREN_CHANGED:
                request.child_watches.append(path)

        yield self.send(request)

    @gen.coroutine
    def close(self):
        self.closing = True

        yield self.send(protocol.CloseRequest())
        self.state.transition_to(States.LOST)

        yield self.conn.close(self.timeout)



import logging
import random

from tornado import gen
from zoonado import exc


log = logging.getLogger(__name__)


def arguments(parser):
    parser.add_argument(
        "--path", "-b", type=str, default="/example-counter",
        help="ZNode path to use for the barrier."
    )
    parser.add_argument(
        "--workers", "-w", type=int, default=5,
        help="Number of worker coroutines."
    )


@gen.coroutine
def run(client, args):
    yield client.start()

    value_history = []

    def callback(new_value):
        value_history.append(new_value)

    watcher = client.recipes.DataWatcher()
    watcher.add_callback(args.path, callback)

    try:
        data = yield client.get_data(args.path)
        log.info("Initial value is %s", data)
    except exc.NoNode:
        log.info("Initial value is blank")

    yield [
        worker(number, client, args)
        for number in range(args.workers)
    ]

    log.info("Value history: %s", " -> ".join(value_history))

    yield client.close()


@gen.coroutine
def worker(number, client, args):
    log.info("[WORKER #%d] Starting up", number)

    counter = client.recipes.Counter(args.path)

    yield counter.start()

    for _ in range(2):
        op = random.choice(["incr", "incr", "decr", "decr", "set"])
        if op == "incr":
            log.info("[WORKER #%d] Incrementing count", number)
            yield counter.incr()
        elif op == "decr":
            log.info("[WORKER #%d] Decrementing count", number)
            yield counter.decr()
        else:
            new_value = random.choice([4, 8, 12, 16])
            log.info("[WORKER #%d] Setting count to %d", number, new_value)
            yield counter.set_value(new_value)

    yield counter.stop()



import logging

from tornado import gen, ioloop


log = logging.getLogger()


def arguments(parser):
    parser.add_argument(
        "--path", "-b", type=str, default="/example-barrier",
        help="ZNode path to use for the barrier."
    )
    parser.add_argument(
        "--workers", "-w", type=int, default=5,
        help="Number of worker coroutines."
    )


@gen.coroutine
def run(client, args):
    yield client.start()

    barrier = client.recipes.Barrier(args.path)

    yield barrier.create()

    for i in range(args.workers):
        yield gen.sleep(1)
        ioloop.IOLoop.current().add_callback(worker, i, client, args.path)

    yield gen.sleep(2)

    yield barrier.lift()

    yield gen.sleep(3)

    yield client.close()


@gen.coroutine
def worker(number, client, barrier_path):
    log.info("[WORKER #%d] Starting up", number)

    barrier = client.recipes.Barrier(barrier_path)

    log.info("[WORKER #%d] Waiting on barrier...", number)
    yield barrier.wait()

    while True:
        log.info("[WORKER #%d] Doing work!", number)
        yield gen.sleep(.3)



import logging
import random

from tornado import gen


log = logging.getLogger()


def arguments(parser):
    parser.add_argument(
        "--workers", "-w", type=int, default=3,
        help="Number of workers to launch."
    )
    parser.add_argument(
        "--lock-path", "-p", type=str, default="examplelock",
        help="ZNode path to use for the lock."
    )


@gen.coroutine
def run(client, args):
    log.info("Launching %d workers.", args.workers)

    yield client.start()

    order = list(range(args.workers))
    random.shuffle(order)

    yield [work(i, client, args) for i in order]

    yield client.close()


@gen.coroutine
def work(number, client, args):
    lock = client.recipes.Lock(args.lock_path)

    num_iterations = 3

    log.info("[WORKER #%d] Acquiring lock...", number)
    with (yield lock.acquire()) as check:
        log.info("[WORKER #%d] Got lock!", number)

        for _ in range(num_iterations):
            wait = random.choice([1, 2, 3])
            if not check():
                log.warn("[WORKER #%d] lost my lock!", number)
                break

            log.info("[WORKER #%d] working %d secs", number, wait)
            yield gen.sleep(wait)

        log.info("[WORKER #%d] Done!", number)



from __future__ import unicode_literals

import logging
import time

from tornado import gen

from zoonado import exc, WatchEvent

from .sequential import SequentialRecipe

log = logging.getLogger(__name__)


class DoubleBarrier(SequentialRecipe):

    def __init__(self, base_path, min_participants):
        super(DoubleBarrier, self).__init__(base_path)
        self.min_participants = min_participants

    @property
    def sentinel_path(self):
        return self.sibling_path("sentinel")

    @gen.coroutine
    def enter(self, timeout=None):
        log.debug("Entering double barrier %s", self.base_path)
        time_limit = None
        if timeout is not None:
            time_limit = time.time() + timeout

        barrier_lifted = self.client.wait_for_event(
            WatchEvent.CREATED, self.sentinel_path
        )
        if time_limit:
            barrier_lifted = gen.with_timeout(time_limit, barrier_lifted)

        exists = yield self.client.exists(path=self.sentinel_path, watch=True)

        yield self.create_unique_znode("worker")

        _, participants = yield self.analyze_siblings()

        if exists:
            return

        elif len(participants) >= self.min_participants:
            yield self.create_znode(self.sentinel_path)
            return

        try:
            yield barrier_lifted
        except gen.TimeoutError:
            raise exc.TimeoutError

    @gen.coroutine
    def leave(self, timeout=None):
        log.debug("Leaving double barrier %s", self.base_path)
        time_limit = None
        if timeout is not None:
            time_limit = time.time() + timeout

        owned_positions, participants = yield self.analyze_siblings()
        while len(participants) > 1:
            if owned_positions["worker"] == 0:
                yield self.wait_on_sibling(participants[-1], time_limit)
            else:
                yield self.delete_unique_znode("worker")
                yield self.wait_on_sibling(participants[0], time_limit)

            owned_positions, participants = yield self.analyze_siblings()

        if len(participants) == 1 and "worker" in owned_positions:
            yield self.delete_unique_znode("worker")
            try:
                yield self.client.delete(self.sentinel_path)
            except exc.NoNode:
                pass



from __future__ import unicode_literals

from tornado import gen

from zoonado import protocol


class Transaction(object):

    def __init__(self, client):
        self.client = client
        self.request = protocol.TransactionRequest()

    def check_version(self, path, version):
        path = self.client.normalize_path(path)

        self.request.add(
            protocol.CheckVersionRequest(path=path, version=version)
        )

    def create(
            self, path, data=None, acl=None,
            ephemeral=False, sequential=False, container=False
    ):
        if container and not self.client.features.containers:
            raise ValueError("Cannot create container, feature unavailable.")

        path = self.client.normalize_path(path)
        acl = acl or self.client.default_acl

        if self.client.features.create_with_stat:
            request_class = protocol.Create2Request
        else:
            request_class = protocol.CreateRequest

        request = request_class(path=path, data=data, acl=acl)
        request.set_flags(ephemeral, sequential, container)

        self.request.add(request)

    def set_data(self, path, data, version=-1):
        path = self.client.normalize_path(path)

        self.request.add(
            protocol.SetDataRequest(path=path, data=data, version=version)
        )

    def delete(self, path, version=-1):
        path = self.client.normalize_path(path)

        self.request.add(
            protocol.DeleteRequest(path=path, version=version)
        )

    @gen.coroutine
    def commit(self):
        if not self.request.requests:
            raise ValueError("No operations to commit.")

        response = yield self.client.send(self.request)
        pairs = zip(self.request.requests, response.responses)

        result = Result()
        for request, reply in pairs:
            if isinstance(reply, protocol.CheckVersionResponse):
                result.checked.add(self.client.denormalize_path(request.path))
            elif isinstance(reply, protocol.CreateResponse):
                result.created.add(self.client.denormalize_path(request.path))
            elif isinstance(reply, protocol.SetDataResponse):
                result.updated.add(self.client.denormalize_path(request.path))
            elif isinstance(reply, protocol.DeleteResponse):
                result.deleted.add(self.client.denormalize_path(request.path))

        raise gen.Return(result)


class Result(object):

    def __init__(self):
        self.checked = set()
        self.created = set()
        self.updated = set()
        self.deleted = set()

    def __bool__(self):
        return sum([
            len(self.checked),
            len(self.created),
            len(self.updated),
            len(self.deleted),
        ]) > 0

    def __nonzero__(self):
        return self.__bool__()



from __future__ import unicode_literals

import logging
import struct

from six import BytesIO

from zoonado import exc

from .request import Request
from .response import Response, response_xref
from .part import Part
from .primitives import Int, Bool


error_struct = struct.Struct("!" + Int.fmt)


log = logging.getLogger(__name__)


class MultiHeader(Part):
    """
    """
    parts = (
        ("type", Int),
        ("done", Bool),
        ("error", Int),
    )


class TransactionRequest(Request):
    """
    """
    opcode = 14

    def __init__(self, *args, **kwargs):
        super(TransactionRequest, self).__init__(*args, **kwargs)
        self.requests = []

    def add(self, request):
        self.requests.append(request)

    def serialize(self, xid=None):
        buff = BytesIO()

        formats = []
        data = []
        if xid is not None:
            formats.append(Int.fmt)
            data.append(xid)
        if self.opcode:
            formats.append(Int.fmt)
            data.append(self.opcode)

        for request in self.requests:
            header = MultiHeader(type=request.opcode, done=False, error=-1)
            header_format, header_data = header.render()
            formats.append(header_format)
            data.extend(header_data)

            payload_format, payload_data = request.render()
            formats.append(payload_format)
            data.extend(payload_data)

        footer = MultiHeader(type=-1, done=True, error=-1)
        footer_format, footer_data = footer.render()
        formats.append(footer_format)
        data.extend(footer_data)

        buff.write(struct.pack("!" + "".join(formats), *data))

        return buff.getvalue()

    def __str__(self):
        return "Txn[%s]" % ", ".join(map(str, self.requests))


class TransactionResponse(Response):
    """
    """
    opcode = 14

    def __init__(self, *args, **kwargs):
        super(TransactionResponse, self).__init__(*args, **kwargs)
        self.responses = []

    @classmethod
    def deserialize(cls, raw_bytes):
        instance = cls()

        header, offset = MultiHeader.parse(raw_bytes, 0)
        while not header.done:
            if header.type == -1:
                error_code = error_struct.unpack_from(raw_bytes, offset)[0]
                offset += error_struct.size
                instance.responses.append(exc.get_response_error(error_code))
                header, offset = MultiHeader.parse(raw_bytes, offset)
                continue

            response_class = response_xref[header.type]

            response, offset = response_class.parse(raw_bytes, offset)
            instance.responses.append(response)

            header, offset = MultiHeader.parse(raw_bytes, offset)

        return instance

    def __str__(self):
        return "Txn[%s]" % ", ".join(map(str, self.responses))



from __future__ import unicode_literals

import logging

import six
from tornado import gen, ioloop

from .children_watcher import ChildrenWatcher
from .data_watcher import DataWatcher
from .recipe import Recipe


log = logging.getLogger(__name__)


class TreeCache(Recipe):

    sub_recipes = {
        "data_watcher": DataWatcher,
        "child_watcher": ChildrenWatcher,
    }

    def __init__(self, base_path, defaults=None):
        super(TreeCache, self).__init__(base_path)
        self.defaults = defaults or {}
        self.root = None

    @gen.coroutine
    def start(self):
        log.debug("Starting znode tree cache at %s", self.base_path)

        self.root = ZNodeCache(
            self.base_path, self.defaults,
            self.client, self.data_watcher, self.child_watcher,
        )

        yield self.ensure_path()

        yield self.root.start()

    def stop(self):
        self.root.stop()

    def __getattr__(self, attribute):
        return getattr(self.root, attribute)

    def as_dict(self):
        return self.root.as_dict()


class ZNodeCache(object):

    def __init__(self, path, defaults, client, data_watcher, child_watcher):
        self.path = path

        self.client = client
        self.defaults = defaults

        self.data_watcher = data_watcher
        self.child_watcher = child_watcher

        self.children = {}
        self.data = None

    @property
    def dot_path(self):
        return self.path[1:].replace("/", ".")

    @property
    def value(self):
        return self.data

    def __getattr__(self, name):
        if name not in self.children:
            raise AttributeError

        return self.children[name]

    @gen.coroutine
    def start(self):
        data, children = yield [
            self.client.get_data(self.path),
            self.client.get_children(self.path)
        ]

        self.data = data
        for child in children:
            self.add_child_znode_cache(child)

        yield [child.start() for child in self.children.values()]

        self.data_watcher.add_callback(self.path, self.data_callback)
        self.child_watcher.add_callback(self.path, self.child_callback)

    def stop(self):
        self.data_watcher.remove_callback(self.path, self.data_callback)
        self.child_watcher.remove_callback(self.path, self.child_callback)

    def child_callback(self, new_children):
        removed_children = set(self.children.keys()) - set(new_children)
        added_children = set(new_children) - set(self.children.keys())

        for removed in removed_children:
            log.debug("Removed child %s", self.dot_path + "." + removed)
            child = self.children.pop(removed)
            child.stop()

        for added in added_children:
            log.debug("added child %s", self.dot_path + "." + added)
            self.add_child_znode_cache(added)
            ioloop.IOLoop.current().add_callback(self.children[added].start)

    def data_callback(self, data):
        log.debug("New value for %s: %r", self.dot_path, data)
        self.data = data

    def add_child_znode_cache(self, child_name):
        self.children[child_name] = ZNodeCache(
            self.path + "/" + child_name, self.defaults.get(child_name, {}),
            self.client, self.data_watcher, self.child_watcher
        )

    def as_dict(self):
        if self.children:
            return {
                child_path: child_znode.as_dict()
                for child_path, child_znode in six.iteritems(self.children)
            }

        return self.data
